from datetime import datetime, timedelta
from enum import Enum

from tic.reputation.shared.dtos.cynic_analysis_result import CynicAnalysisResult
from tic.shared.dto import Dto
from tic.shared.utils.date_util import (
    epoch_time_to_utc_datetime,
    utc_datetime_to_epoch_time,
    utc_now,
    to_iso_8601,
    match_str_strptime_to_datetime,
)


class CynicVerdict(Enum):
    UNKNOWN = -1
    MALWARE = 0
    CLEAN = 1


class Cynic(Dto):
    decode_settings = {
        "verdict": lambda x: None if x is None else CynicVerdict(x),
        "ttl": lambda x: epoch_time_to_utc_datetime(utc_now())
        if x is None
        else epoch_time_to_utc_datetime(x),
        "analysis_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
    }

    encode_settings = {
        "verdict": lambda x: None if x is None else x.value,
        "ttl": lambda x: utc_datetime_to_epoch_time(utc_now())
        if x is None
        else utc_datetime_to_epoch_time(x),
        "analysis_date": lambda x: to_iso_8601(x),
    }

    def __init__(
        self,
        verdict: CynicVerdict = CynicVerdict.UNKNOWN,
        analysis_date: datetime = None,
        threat_name: str = None,
        status: str = None,
        ttl: datetime = None,  # Absolute timestamp at which this APC data expires.
    ):
        self.verdict: CynicVerdict = verdict
        self.analysis_date: analysis_date = (
            utc_now() if analysis_date is None else analysis_date
        )
        self.threat_name: str = threat_name
        self.status: str = status
        self.ttl: datetime = ttl if ttl is not None else utc_now()

    @classmethod
    def from_cynic_analysis_result(cls, result: CynicAnalysisResult):
        verdict = (
            CynicVerdict.MALWARE
            if result.verdict and result.verdict.lower() == "malware"
            else CynicVerdict.CLEAN
        )
        return Cynic(
            verdict=verdict,
            analysis_date=datetime(1970, 1, 1)
            + timedelta(milliseconds=result.verdictTime),
            threat_name=result.threatName,
            status=result.status,
        )

    def waiting_for_fast_result(self) -> bool:
        """
        Returns true if waiting for cynic result for less than
        """
        if self.status and self.status == "UPLOADED":
            if (
                self.analysis_date
                and (utc_now() - self.analysis_date).total_seconds() < 120
            ):
                return True
        return False
