from datetime import timedelta
from typing import Tuple
import json
import random

from .cynic import Cynic
from ..enums.reputation_category import ReputationCategory
from ..enums.reputation_channel import ReputationChannel
from ..enums.reputation_status import ReputationStatus
from ..enums.reputation_tags import ReputationTags
from ..enums.signature_status import SignatureStatus

from ....shared.dto import Dto
from ....shared.utils.date_util import (
    datetime_to_str,
    match_str_strptime_to_datetime,
    to_iso_8601,
    utc_datetime_to_epoch_time,
    utc_now,
)


class ReputationCacheItem(Dto):
    decode_settings = {
        "date_created": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "date_modified": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "ttl": lambda x: utc_now() if x is None else match_str_strptime_to_datetime(x),
        "reputation": lambda x: Reputation()
        if x is None
        else Reputation.from_dict(x, True, True),
        "scan_info": lambda x: None if x is None else ScanInfo.from_dict(x, True, True),
        "signature_info": lambda x: None
        if x is None
        else SignatureInfo.from_dict(x, True, True),
        "status": lambda x: ReputationStatus.NotFound
        if x is None
        else ReputationStatus(x),
        "channel": lambda x: ReputationChannel.STABLE
        if x is None
        else ReputationChannel(x),
        "version": lambda x: tuple(int(y) for y in x.split(".")),
    }

    encode_settings = {
        "date_created": lambda x: to_iso_8601(x),
        "date_modified": lambda x: to_iso_8601(x),
        "ttl": lambda x: to_iso_8601(x),
        "status": lambda x: ReputationStatus.NotFound.value if x is None else x.value,
        "apc_ttl": lambda x: None if x is None else to_iso_8601(x),
        "channel": lambda x: ReputationChannel.STABLE.value if x is None else x.value,
        "version": lambda x: None if x is None else ".".join((str(y) for y in x)),
    }

    def __init__(
        self,
        sha256=None,
        date_modified=None,
        file_name=None,
        reputation=None,
        scan_info=None,
        signature_info=None,
        status=None,
        is_interesting=False,
        file_exists=False,
        channel: ReputationChannel = None,
        version: Tuple[int] = None,
    ):
        self.date_created = utc_now()
        self.date_modified = date_modified if date_modified is not None else utc_now()
        self.sha256 = sha256
        self.file_name = file_name
        self.set_need_file(is_interesting, file_exists)
        self.reputation = reputation if reputation is not None else Reputation()
        self.channel = ReputationChannel.STABLE if channel is None else channel
        self.scan_info = scan_info if scan_info is not None else ScanInfo()
        self.signature_info = signature_info
        self.status = status if status is not None else ReputationStatus.NotFound
        self.version: Tuple[int] = (1, 0) if version is None else version
        self.set_ttl(is_interesting)

    def set_need_file(
        self,
        is_interesting: bool = False,
        file_exists: bool = False,
        definitively_categorized: bool = False,
    ):
        """
        Sets the need_file flag for the response
        :param is_interesting: is_interesting flag for this hash
        :param file_exists: Flag representing whether or not we have the file
        :param apc_upload_allowed: Flag indicating whether or not apc upload is allowed for hash
        :param definitively_categorized: Reputation item is definitively categorized
        """
        self.need_file = (
            True
            if is_interesting and not definitively_categorized and not file_exists
            else False
        )

    def set_ttl(self, is_interesting=False, cynic: Cynic = None):
        """
        Set the TTL for the CacheItem.
        Returns the relative TTL.
        """
        now = utc_now()  # Use same definition of NOW
        if self.status == ReputationStatus.Submitted:
            self.ttl = now
            self.ttl_rel = 0
            return

        # check whether reputation has resolved APC status:
        """
        _needs_callback = not self.definitive_verdict() and (
            (
                is_interesting
                and (
                    apc is None
                    or apc.status == ApcStatus.UPLOAD
                    or (
                        apc.status == ApcStatus.OK
                        and apc.category == ApcCategory.UNKNOWN
                    )
                    or (not has_file or not apc_upload_allowed)
                )
            )
            or (apc is not None and apc.status == ApcStatus.IN_PROGRESS)
        )
        """

        _needs_callback = not self.definitive_verdict() and (
            (
                is_interesting
                and (
                    cynic is None  # No file yet
                    or (
                        cynic and cynic.waiting_for_fast_result()
                    )  # File is uploaded to cynic
                )
            )
        )

        category = self.reputation.category
        if self.reputation.locked:  # Rep isn't going to change, so make TTL super long
            delta = timedelta(days=90 + 7 * random.random())
        elif _needs_callback:
            delta = timedelta(seconds=2)
        elif (
            category == ReputationCategory.TRUSTED_WHITE_LIST
        ):  # Rep very unlikely to change
            delta = timedelta(days=60 + 7 * random.random())
        elif category == ReputationCategory.COMMON_WHITE_LIST:  # Rep unlikely to change
            delta = timedelta(days=30 + 7 * random.random())
        elif category == ReputationCategory.NOT_LISTED:
            delta = timedelta(days=14 + 7 * random.random())
        else:  # KNOWN_MALWARE, SUSPECT_MALWARE, PUP, ADAPTIVE_WHITE_LIST
            age = now - self.scan_info.first_seen
            if age.days >= 30:  # Hash is old, so less likely to change
                delta = timedelta(days=30 + 7 * random.random())
            else:  # Hash is new, so be diligent
                delta = timedelta(hours=24 + 12 * random.random())

        self.ttl = now + delta
        self.ttl_rel = delta.total_seconds()

    def definitive_verdict(self):
        return self.reputation.locked or self.reputation.category in {
            ReputationCategory.KNOWN_MALWARE,
            ReputationCategory.TRUSTED_WHITE_LIST,
        }

    def to_response(self, as_dict=False):
        d = self.to_dict(
            True, True, key_ignorelist={"apc_ttl", "metadata", "is_interesting"}
        )

        if "max_file_size" in d and d.get("need_file", False) is not True:
            del d["max_file_size"]

        # legacy field naming
        d["hash"] = self.sha256
        del d["sha256"]

        # duplicate fields
        if self.scan_info is not None:
            d["scan_info"]["sha256"] = self.sha256

        # lists
        if self.reputation is not None:
            if self.reputation.comment is not None:
                d["reputation"]["comment"] = "; ".join(
                    {tag.name for tag in self.reputation.comment}
                )
            if self.reputation.engine is not None:
                d["reputation"]["engine"] = ", ".join(self.reputation.engine)

        # enums
        d["status"] = self.status.name
        d["reputation"]["category"] = (
            self.reputation.category.name
            if self.reputation.category != ReputationCategory.NOT_LISTED
            else None
        )
        if (
            self.signature_info is not None
            and self.signature_info.signature_status is not None
        ):
            d["signature_info"][
                "signature_status"
            ] = self.signature_info.signature_status.name

        def encode_datetime(x):
            return datetime_to_str(x, "%Y-%m-%d %H:%M:%S.%f")

        # dates
        d["date_created"] = encode_datetime(self.date_created)
        d["date_modified"] = encode_datetime(self.date_modified)
        if self.scan_info is not None and self.scan_info.first_seen is not None:
            d["scan_info"]["first_seen"] = encode_datetime(self.scan_info.first_seen)

        # ttl still in epoch time
        d["ttl"] = utc_datetime_to_epoch_time(self.ttl)
        d["ttl_ms"] = int(d["ttl"] * 1000)
        d["ttl_rel"] = int(self.ttl_rel * 1000)  # Relative time, TTL in ms

        # handle removing MFG data if not malicious.
        # match what is returned for NOT_LISTED.
        if self.reputation.category not in {
            ReputationCategory.KNOWN_MALWARE,
            ReputationCategory.SUSPECT_MALWARE,
            ReputationCategory.PUP,
        }:
            d["reputation"]["av_sub_category"] = None
            d["reputation"]["virus_name"] = None
            d["reputation"]["engine"] = ""

        return (
            d
            if as_dict is True
            else json.dumps(
                d, separators=(",", ":"), sort_keys=True, ensure_ascii=False
            )
        )

    def to_json(
        self, pretty=False, minimize=False, key_ignorelist=None, ensure_ascii=False
    ):
        if key_ignorelist is None:
            key_ignorelist = set()
        key_ignorelist.add("ttl")

        return super().to_json(
            pretty=pretty,
            minimize=minimize,
            key_ignorelist=key_ignorelist,
            ensure_ascii=ensure_ascii,
        )


class Reputation(Dto):
    decode_settings = {
        "category": lambda x: ReputationCategory.NOT_LISTED
        if x is None
        else ReputationCategory(x),
        "comment": lambda x: list() if x is None else [ReputationTags(y) for y in x],
    }
    encode_settings = {
        "category": lambda x: ReputationCategory.NOT_LISTED.value
        if x is None
        else x.value,
        "comment": lambda x: None if x is None or len(x) == 0 else [y.value for y in x],
    }

    def __init__(
        self,
        av_sub_category=None,
        category=None,
        comment=None,
        engine=None,
        locked=False,
        virus_name=None,
    ):
        self.av_sub_category = av_sub_category
        self.category = (
            category if category is not None else ReputationCategory.NOT_LISTED
        )
        self.comment = comment if comment is not None else list()
        self.engine = engine if engine is not None else list()
        self.locked = locked
        self.virus_name = virus_name


class ScanInfo(Dto):
    decode_settings = {
        "first_seen": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x)
    }

    encode_settings = {"first_seen": lambda x: to_iso_8601(x)}

    def __init__(
        self, md5=None, sha1=None, first_seen=None, positives=0, total_scans=0
    ):
        self.md5 = md5
        self.sha1 = sha1
        self.first_seen = first_seen if first_seen is not None else utc_now()
        self.positives = positives
        self.total_scans = total_scans


class SignatureInfo(Dto):
    decode_settings = {
        "signature_status": lambda x: SignatureStatus.UNKNOWN
        if x is None
        else SignatureStatus(x)
    }

    encode_settings = {
        "signature_status": lambda x: SignatureStatus.UNKNOWN.value
        if x is None
        else x.value
    }

    def __init__(
        self,
        publisher_name=None,
        signature_status=None,
        signed_by=None,
        signing_authority=None,
    ):
        self.publisher_name = publisher_name
        self.signature_status = (
            signature_status
            if signature_status is not None
            else SignatureStatus.UNKNOWN
        )
        self.signed_by = signed_by
        self.signing_authority = signing_authority
