from ..dtos.reputation_cache_item import ReputationCacheItem
from ..dtos.reputation_cache_item import Reputation as CacheReputation
from ..dtos.reputation_cache_item import SignatureInfo as CacheSignatureInfo
from ..dtos.reputation_cache_item import ScanInfo as CacheScanInfo
from .apc import Apc
from .candidate import Candidate
from .cynic import Cynic, CynicVerdict

from ..enums.mfg_classification_description import MfgClassificationDescription
from ..enums.reputation_category import ReputationCategory
from ..enums.reputation_status import ReputationStatus
from ..enums.reputation_tags import ReputationTags
from ..enums.signature_status import SignatureStatus

from ..dtos.reff import ReffReputation
from ....shared.apc.enums.apc_category import ApcCategory
from ....shared.apc.enums.apc_status import ApcStatus
from ....shared.apc.enums.apc_verdict import ApcVerdict
from ....shared.aws.dynamodb import DynamoDbModel
from ....shared.dto import Dto
from ....shared.utils.date_util import (
    utc_now,
    match_str_strptime_to_datetime,
    to_iso_8601,
)


class ReputationItem(DynamoDbModel):
    encode_names = {"apc": "apcv4"}
    decode_names = {"apcv4": "apc"}
    decode_settings = {
        "cb_first_seen": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "feed_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "categorization_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "date_modified": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "reputation": lambda x: Reputation()
        if x is None
        else Reputation.from_dict(x, True, True),
        "scan_details": lambda x: None
        if x is None
        else ScanDetails.from_dict(x, True, True),
        "mfg": lambda x: None if x is None else Mfg.from_dict(x, True, True),
        "signature_info": lambda x: None
        if x is None
        else SignatureInfo.from_dict(x, True, True),
        "sources": lambda x: list() if x is None else x,
        "apc": lambda x: None if x is None else Apc.from_dict(x, True, True),
        "cynic": lambda x: None if x is None else Cynic.from_dict(x, True, True),
        "candidate": lambda x: None
        if x is None
        else Candidate.from_dict(x, True, True),
        "reff": lambda x: ReffReputation()
        if x is None
        else ReffReputation.from_dict(x, True, True),
        "cdc_reputation": lambda x: None
        if x is None
        else Reputation.from_dict(x, True, True),
    }

    encode_settings = {
        "cb_first_seen": lambda x: to_iso_8601(x),
        "feed_date": lambda x: to_iso_8601(x),
        "categorization_date": lambda x: to_iso_8601(x),
        "date_modified": lambda x: to_iso_8601(x),
        "sources": lambda x: None if x is None or len(x) == 0 else x,
    }

    pkey = {"sha256": "S"}

    def __init__(
        self,
        sha256=None,
        md5=None,
        sha1=None,
        tlsh=None,
        sources=None,
        reputation=None,
        file_name=None,
        cb_first_seen=None,
        feed_date=None,
        categorization_date=None,
        date_modified=None,
        scan_details=None,
        mfg=None,
        signature_info=None,
        apc: Apc = None,
        candidate=None,
        version=0,
        cynic: Cynic = None,
        reff=None,
        cdc_reputation=None,
    ):
        self.sha256 = sha256
        self.md5 = md5
        self.sha1 = sha1
        self.tlsh = tlsh
        self.sources = sources if sources is not None else list()
        self.reputation = Reputation() if reputation is None else reputation
        self.file_name = file_name
        self.cb_first_seen = utc_now() if cb_first_seen is None else cb_first_seen
        self.feed_date = utc_now() if feed_date is None else feed_date
        self.categorization_date = (
            utc_now() if categorization_date is None else categorization_date
        )
        self.date_modified = utc_now() if date_modified is None else date_modified
        self.scan_details = scan_details
        self.mfg = mfg
        self.signature_info = signature_info
        self.apc: Apc = apc
        self.candidate = candidate
        self.version = version
        self.cynic = cynic
        self.reff = reff
        self.cdc_reputation = cdc_reputation

    def is_definitively_categorized(self):
        return self.reputation.is_locked() or self.reputation.category in {
            ReputationCategory.TRUSTED_WHITE_LIST,
            ReputationCategory.KNOWN_MALWARE,
        }

    @property
    def _classification_scanners_count(self):
        if self.mfg is None:
            return 0
        # TODO Temporary fix since classification isn't going to be filled until MFG back fill is complete (TPLAT-6233)
        elif (
            self.mfg.classification_scanners is None
            or self.mfg.classification_scanners == []
        ):
            return max(
                self.mfg.malware_family.num_scanner_votes,
                self.mfg.malware_group.num_scanner_votes,
                self.mfg.platform.num_scanner_votes,
            )

        return len(set(self.mfg.classification_scanners))

    def categorize(self):
        # we can assume here that there will always be either CDCR, MFG, or APC data because of the new model--
        # should never get forwarded truly empty data

        # Safeguard here in case defaults do not work for some reason.
        if self.reputation is None:
            self.reputation = Reputation()

        # make reputation.comment a set so we can add and remove items easily
        # will cast back to list at the end of categorization
        if self.reputation.comment is None:
            self.reputation.comment = set()
        else:
            self.reputation.comment = set(self.reputation.comment)

        if self.cynic:
            self._perform_cynic_adjustments()

        # This won't be executed under normal processing
        if self.reputation.category == ReputationCategory.NOT_LISTED:
            self.reputation.trust = 0
            self.reputation.threat = 0
            self.reputation.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            self.reputation.comment.add(ReputationTags.REFF_INCONCLUSIVE)
        self.categorization_date = utc_now()
        self.reputation.comment = list(self.reputation.comment)

        # Prevoius logic
        # self._categorize_trust_threat()
        #
        # # Add locked tag whether we need to use MFG to update category or not.
        # # May not be used for category since MFG can't be locked without rep also locked.
        # if self.mfg is not None and self.mfg.locked is True:
        #     self.reputation.comment.add(ReputationTags.MFG_LOCKED)
        #
        # # check for whether we need MFG added
        # if (
        #         self.reputation.is_locked()
        #         or self.reputation.category == ReputationCategory.TRUSTED_WHITE_LIST
        # ):
        #     self.reputation.comment.add(ReputationTags.DEFINITIVE_BEFORE_MFG)
        #     return
        # if self.mfg is not None:
        #     self.reputation.comment.add(ReputationTags.MFG_DATA_EVALUATED)
        #     self._categorize_with_mfg()
        #
        # self.reputation.comment = list(self.reputation.comment)
        # if self.apc is not None:
        #     self._perform_apc_adjustments()
        # if self.cynic is not None:
        #     self._perform_cynic_adjustments()
        # self.categorization_date = utc_now()

    def _common_white_update(self):
        """
        updates category from ADAPTIVE to COMMON if no positives and first seen > 30 days or is validly signed.
        """
        if (
            self.scan_details is None
            or self.signature_info is None
            or self.cb_first_seen is None
        ):
            return
        elif self.scan_details.positives == 0 and self.scan_details.total_scans > 0:
            first_seen = (
                self.scan_details.first_scanned_date
                if self.scan_details.first_scanned_date < self.cb_first_seen
                else self.cb_first_seen
            )
            if (
                self.signature_info.signature_status
                == SignatureStatus.SIGNED_AND_VERIFIED
                or (utc_now() - first_seen).days >= 30
            ):
                self.reputation.category = ReputationCategory.COMMON_WHITE_LIST
                self.reputation.comment.add(ReputationTags.COMMON_WHITE_CRITERIA_MET)
        return

    def _categorize_trust_threat(self):
        if self.reputation.threat == 100:
            self.reputation.category = ReputationCategory.KNOWN_MALWARE
            self.reputation.comment.add(ReputationTags.CB_HIGH_THREAT)
        elif self.reputation.threat >= 50:
            self.reputation.category = ReputationCategory.PUP
            self.reputation.comment.add(ReputationTags.CB_PUP)
        else:
            self.reputation.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            if (
                self.scan_details is not None
                and self.scan_details.positives is not None
                and not self.reputation.is_locked()
            ):
                self.reputation.comment.add(ReputationTags.CB_SCORES_INCONCLUSIVE)
                if self.scan_details.positives >= 6:
                    self.reputation.comment.add(ReputationTags.SUSPECT_FROM_POSITIVES)
                    self.reputation.category = ReputationCategory.SUSPECT_MALWARE
                else:
                    self.reputation.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            if self.reputation.category == ReputationCategory.ADAPTIVE_WHITE_LIST:
                self._common_white_update()

        # handles when only trust is locked or when threat locked low and trust high
        if self.reputation.trust >= 7:
            self.reputation.category = ReputationCategory.TRUSTED_WHITE_LIST
            self.reputation.comment.add(ReputationTags.CB_HIGH_TRUST)

    def _categorize_with_mfg(self):
        # Relying on MFG locked only if reputation locked-
        # We should already have category based on rep lock.

        # add US_CERT tag if UsCert in mfg scanners.
        if "UsCert" in self.mfg.classification_scanners:
            self.reputation.comment.add(ReputationTags.US_CERT_MALWARE)

        # use MFG PUP or malware_family of Mimikatz or LaZange to classify
        if not self.reputation.is_locked():
            if self.reff and self.reff.cb_classification:
                # use cb_classification if reputation score is not part of 119, -119, -61, -69 bands or if it is null
                if (
                    not self.reff.reputation_score
                    or self.reff.reputation_score not in (119, -119, -61, -69)
                ):
                    if self.reff.cb_classification == "PUP":
                        self.reputation.category = ReputationCategory.PUP
                        self.reputation.comment.add(
                            ReputationTags.REFF_CB_CLASSIFICATION
                        )
                        return
                    elif (
                        self.reff.reputation_score and self.reff.reputation_score < 52
                    ) or self.reff.reputation_score is None:
                        if self.reff.cb_classification == "KNOWN_MALWARE":
                            self.reputation.category = ReputationCategory.KNOWN_MALWARE
                            self.reputation.comment.add(
                                ReputationTags.REFF_CB_CLASSIFICATION
                            )
                            return
                        elif self.reff.cb_classification == "SUSPECT_MALWARE":
                            self.reputation.category = (
                                ReputationCategory.SUSPECT_MALWARE
                            )
                            self.reputation.comment.add(
                                ReputationTags.REFF_CB_CLASSIFICATION
                            )
                            return

            if (
                self.reff
                and self.reff.reputation_score
                and (
                    self.reff.reputation_score == -61
                    or (
                        -64 < self.reff.reputation_score <= -25
                        and self.mfg.malware_group
                        and self.mfg.malware_group.result
                        and self.mfg.malware_group.result.startswith("PUA")
                    )
                )
            ):
                self.reputation.category = ReputationCategory.PUP
                self.reputation.comment.add(ReputationTags.MFG_PUP)
                return

            elif (
                self._classification_scanners_count > 2
                or "Avira" in self.mfg.classification_scanners
            ):
                # Classify as malware if malware_family is Mimikatz or LaZange
                if (
                    self.mfg.malware_family.result == "Mimikatz"
                    or self.mfg.malware_family.result == "LaZange"
                ):
                    # Update the signature and malware group
                    self.mfg.malware_group.result = "PWS"
                    self.mfg.signature = (
                        "PWS:"
                        + (self.mfg.platform.result if self.mfg.platform.result else "")
                        + "/"
                        + self.mfg.malware_family.result
                        + (("." + ".".join(self.mfg.tags)) if self.mfg.tags else "")
                    )
                    if "Avira" not in self.mfg.classification_scanners:
                        self.reputation.category = ReputationCategory.SUSPECT_MALWARE
                        self.reputation.comment.add(ReputationTags.MFG_SUSPECT)
                        return
                    else:
                        self.reputation.category = ReputationCategory.KNOWN_MALWARE
                        self.reputation.comment.add(ReputationTags.MFG_AVIRA_CONVICTION)
                        return
                # use MFG PUP if caught by MFG
                elif self.mfg.malware_group.result in {"Adware", "PUP"}:
                    self.reputation.category = ReputationCategory.PUP
                    self.reputation.comment.add(ReputationTags.MFG_PUP)
                    return

        if not self.is_definitively_categorized():
            if (
                self._classification_scanners_count > 2
                or "Avira" in self.mfg.classification_scanners
            ):
                if "Avira" not in self.mfg.classification_scanners:
                    self.reputation.category = ReputationCategory.SUSPECT_MALWARE
                    self.reputation.comment.add(ReputationTags.MFG_SUSPECT)
                    return
                else:
                    self.reputation.category = ReputationCategory.KNOWN_MALWARE
                    self.reputation.comment.add(ReputationTags.MFG_AVIRA_CONVICTION)
                    return
            # no data to convict in MFG
            self.reputation.comment.add(ReputationTags.MFG_NO_CONVICTION)
            # don't downgrade suspect, pup classifications from trust/threat categorization
            if self.reputation.category not in {
                ReputationCategory.SUSPECT_MALWARE,
                ReputationCategory.PUP,
            }:
                self.reputation.category = ReputationCategory.ADAPTIVE_WHITE_LIST
                self._common_white_update()

    def _perform_apc_adjustments(self):
        self.reputation.comment = set(self.reputation.comment)
        self.reputation.comment.add(ReputationTags.APC_DATA_EVALUATED)

        _used_apc_data = False

        # check for already counted avira data in MFG
        if self.mfg and "Avira" in self.mfg.classification_scanners:
            self.reputation.comment.add(ReputationTags.APC_AVIRA_ALREADY_IN_MFG)
            self.reputation.comment = list(self.reputation.comment)
            return

        # Check if more recent Avira data than APC data (if there's no query data use Avira data)
        if "avira" in self.sources and (
            not self.apc.query_date or self.feed_date > self.apc.query_date
        ):
            self.reputation.comment.add(ReputationTags.APC_AVIRA_ALREADY_IN_MFG)
            self.reputation.comment = list(self.reputation.comment)
            return

        # check that avira data is ready
        if self.apc.status != ApcStatus.OK:
            self.reputation.comment.add(ReputationTags.APC_UNREADY)
            self.reputation.comment = list(self.reputation.comment)
            return

        # check that avira has a result
        if self.apc.category == ApcCategory.UNKNOWN:
            self.reputation.comment.add(ReputationTags.APC_UNKNOWN)
            self.reputation.comment = list(self.reputation.comment)
            return

        # upgrade categorization
        if (
            self.apc.verdict == ApcVerdict.PUP
            and self.reputation.category != ReputationCategory.SUSPECT_MALWARE
        ):
            _used_apc_data = True
            self.reputation.category = ReputationCategory.PUP
            self.reputation.comment.add(ReputationTags.APC_PUP)
        elif self.apc.verdict == ApcVerdict.MALWARE:
            _used_apc_data = True
            self.reputation.category = ReputationCategory.KNOWN_MALWARE
            self.reputation.comment.add(ReputationTags.APC_MALICIOUS)

        if _used_apc_data:
            self.reputation.comment.add(ReputationTags.APC_USED)
        else:
            if self.reputation.category == ReputationCategory.NOT_LISTED:
                self.reputation.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            self.reputation.comment.add(ReputationTags.APC_UNUSED)

        self.reputation.comment = list(self.reputation.comment)

    def _perform_cynic_adjustments(self):
        self.reputation.comment = set(self.reputation.comment)
        self.reputation.comment.add(ReputationTags.CYNIC_DATA_EVALUATED)

        # ONLY UPDATE if cynic data comes before the feed
        if self.feed_date <= self.cynic.analysis_date:
            if not self.reputation.is_locked():
                if self.cynic.verdict == CynicVerdict.MALWARE:
                    self.reputation.category = ReputationCategory.KNOWN_MALWARE
                    self.reputation.threat = 100
                    self.reputation.trust = 0
                else:
                    self.reputation.category = ReputationCategory.ADAPTIVE_WHITE_LIST
                    self.reputation.threat = 0
                    self.reputation.trust = 0

    def to_reputation_cache_item(self, is_interesting=False, candidate=False):
        status = ReputationStatus.NotFound
        category = self.reputation.category
        comment = self.reputation.comment
        if category != ReputationCategory.NOT_LISTED:
            status = ReputationStatus.Found
        if candidate and self.candidate is not None:
            status = ReputationStatus.Found
            category = self.candidate.category
            comment.extend(self.candidate.additional_tags)

        return ReputationCacheItem(
            sha256=self.sha256,
            date_modified=self.date_modified,
            file_name=self.file_name,
            status=status,
            reputation=CacheReputation(
                av_sub_category=self.mfg.malware_group.result
                if self.mfg is not None
                else None,
                category=category,
                comment=comment,
                engine=self.mfg.classification_scanners
                if self.mfg is not None
                else list(),
                locked=(self.reputation.threat_locked or self.reputation.trust_locked),
                virus_name=self.mfg.malware_family.result
                if self.mfg is not None
                else None,
            ),
            scan_info=CacheScanInfo(
                md5=self.md5,
                sha1=self.sha1,
                positives=self.scan_details.positives
                if self.scan_details is not None
                else None,
                total_scans=self.scan_details.total_scans
                if self.scan_details is not None
                else None,
                first_seen=(
                    self.cb_first_seen
                    if (
                        self.scan_details is None
                        or self.cb_first_seen < self.scan_details.first_scanned_date
                    )
                    else self.scan_details.first_scanned_date
                ),
            ),
            signature_info=self.signature_info.to_cache_response()
            if self.signature_info is not None
            else None,
            is_interesting=is_interesting,
        )


class Reputation(Dto):
    decode_settings = {
        "category": lambda x: ReputationCategory.NOT_LISTED
        if x is None
        else ReputationCategory(x),
        "comment": lambda x: list() if x is None else [ReputationTags(y) for y in x],
    }
    encode_settings = {
        "category": lambda x: ReputationCategory.NOT_LISTED.value
        if x is None
        else x.value,
        "comment": lambda x: None if x is None or len(x) == 0 else [y.value for y in x],
    }

    def __init__(
        self,
        trust=0,
        threat=0,
        trust_locked=False,
        threat_locked=False,
        category=ReputationCategory.NOT_LISTED,
        comment=None,
    ):
        self.trust = trust
        self.threat = threat
        self.trust_locked = trust_locked
        self.threat_locked = threat_locked
        self.category = category
        self.comment = comment if comment is not None else list()

    def is_locked(self):
        return self.trust_locked or self.threat_locked


class ScanDetails(Dto):
    decode_settings = {
        "first_scanned_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "last_scanned_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
    }

    encode_settings = {
        "first_scanned_date": lambda x: to_iso_8601(x),
        "last_scanned_date": lambda x: to_iso_8601(x),
        "positive_scanners": lambda x: None if x is None or len(x) == 0 else x,
    }

    def __init__(
        self,
        positives=0,
        total_scans=0,
        first_scanned_date=None,
        last_scanned_date=None,
        positive_scanners=None,
    ):
        self.positives = positives
        self.total_scans = total_scans
        self.first_scanned_date = (
            utc_now() if first_scanned_date is None else first_scanned_date
        )
        self.last_scanned_date = (
            utc_now() if last_scanned_date is None else last_scanned_date
        )
        self.positive_scanners = (
            positive_scanners if positive_scanners is not None else list()
        )


class Mfg(Dto):
    decode_settings = {
        "platform": lambda x: None
        if x is None
        else MfgClassificationDetail.from_dict(x, True, True),
        "malware_family": lambda x: None
        if x is None
        else MfgClassificationDetail.from_dict(x, True, True),
        "malware_group": lambda x: None
        if x is None
        else MfgClassificationDetail.from_dict(x, True, True),
        "classification_scanners": lambda x: list() if x is None else x,
        "tags": lambda x: list() if x is None else x,
    }
    encode_settings = {
        "classification_scanners": lambda x: None if x is None or len(x) == 0 else x,
        "tags": lambda x: None if x is None or len(x) == 0 else x,
    }

    def __init__(
        self,
        signature=None,
        tags=None,
        classification_scanners=None,
        platform=None,
        malware_family=None,
        malware_group=None,
        locked=False,
    ):
        self.signature = signature
        self.tags = tags if tags is not None else list()
        self.classification_scanners = (
            classification_scanners if classification_scanners is not None else list()
        )
        self.platform = MfgClassificationDetail() if platform is None else platform
        self.malware_family = (
            MfgClassificationDetail() if malware_family is None else malware_family
        )
        self.malware_group = (
            MfgClassificationDetail() if malware_group is None else malware_group
        )
        self.locked = bool(locked)


class MfgClassificationDetail(Dto):
    decode_settings = {
        "description": lambda x: MfgClassificationDescription.Unset
        if x is None
        else MfgClassificationDescription(x)
    }

    encode_settings = {
        "description": lambda x: MfgClassificationDescription.Unset.value
        if x is None
        else x.value
    }

    def __init__(
        self,
        num_scanner_votes=0,
        result=None,
        description=MfgClassificationDescription.Unset,
    ):
        self.num_scanner_votes = num_scanner_votes
        self.result = result
        self.description = description


class SignatureInfo(Dto):
    decode_settings = {
        "signature_status": lambda x: SignatureStatus.UNKNOWN
        if x is None
        else SignatureStatus(x)
    }
    encode_settings = {
        "signature_status": lambda x: SignatureStatus.UNKNOWN.value
        if x is None
        else x.value
    }

    def __init__(
        self, signed_by=None, authority=None, signature_status=SignatureStatus.UNKNOWN
    ):
        self.signed_by = signed_by
        self.authority = authority
        self.signature_status = signature_status

    def to_cache_response(self):
        if self.signed_by is None and self.signature_status is SignatureStatus.UNKNOWN:
            return None

        return CacheSignatureInfo(
            signature_status=self.signature_status,
            signed_by=self.signed_by,
            publisher_name=self.signed_by,
            signing_authority=self.authority,
        )
