from datetime import datetime, timedelta
from enum import IntEnum
from random import random
from typing import List, Optional

from .error_responses import ReputationErrorResponse
from tic.reputation.shared.dtos.reputation_item import (
    ReputationItem,
    SignatureInfo as DbSignatureInfo,
)
from tic.reputation.shared.enums.reputation_tags import ReputationTags
from tic.reputation.shared.enums.signature_status import SignatureStatus
from tic.shared.apc.enums.apc_category import ApcCategory
from tic.shared.apc.enums.apc_status import ApcStatus
from tic.shared.apc.enums.apc_verdict import ApcVerdict
from tic.shared.dto import Dto
from tic.shared.utils.date_util import utc_now, from_iso_8601, to_iso_8601


class ReputationVerdict(IntEnum):
    TRUSTED = 0
    SOMEWHAT_TRUSTED = 1
    INCONCLUSIVE = 2
    SUSPECT_MALWARE = 3
    MALWARE = 4


class ReputationCategory(IntEnum):
    GENERIC = 0
    TROJAN = 1
    KEYLOGGER = 2
    RANSOMWARE = 3
    ADWARE = 4
    FAKE_AV = 5
    PUP = 6
    MINER = 7
    INFORMATION_STEALER = 8

    @staticmethod
    def from_mfg(malware_group: str = None):
        mapping = {
            ReputationCategory.TROJAN: [
                "trojan",
                "backdoor",
                "dropper",
                "virus",
                "rootkit",
                "malware",
                "remoteadmin",
                "worm",
                "downloader",
                "remoteaccess",
                "rat",
            ],
            ReputationCategory.KEYLOGGER: ["keylog", "keylogger"],
            ReputationCategory.RANSOMWARE: ["ransomware", "lock", "kryptic"],
            ReputationCategory.ADWARE: ["adware", "adtool", "clicker", "adwarensis"],
            ReputationCategory.FAKE_AV: ["fakeav", "fake", "misleading"],
            ReputationCategory.PUP: ["hacktool", "nettool", "pup", "cracktool"],
            ReputationCategory.MINER: ["miner"],
            ReputationCategory.INFORMATION_STEALER: [
                "pws",
                "spyware",
                "banker",
                "gamethief",
            ],
        }
        if isinstance(malware_group, str):
            malware_group = malware_group.lower()
        for category, values in mapping.items():
            if malware_group in values:
                return category
        return ReputationCategory.GENERIC


class CertificateStatus(IntEnum):
    SIGNED = 0
    UNSIGNED = 1
    EXPIRED = 2
    REVOKED = 3


class MalwareInfo(Dto):
    def __init__(self, signature: str = None, group: str = None, family: str = None):
        self.signature: str = signature
        self.group: str = group
        self.family: str = family

    def __bool__(self):
        return bool(self.signature or self.group or self.family)


class Classification(Dto):
    encode_settings = {
        "verdict": lambda x: x.name,
        "category": lambda x: None if not x else x.name,
        "malware_info": lambda x: None if not x else x.to_dict(True, True),
        "verdict_details": lambda x: None if not x else [y.name for y in x],
    }
    decode_settings = {
        "verdict": lambda x: ReputationVerdict.INCONCLUSIVE
        if x is None
        else ReputationVerdict[x],
        "category": lambda x: None if x is None else ReputationCategory[x],
        "malware_info": lambda x: MalwareInfo()
        if x is None
        else MalwareInfo.from_dict(x, True, True),
        "verdict_details": lambda x: list()
        if not x
        else [ReputationTags[y] for y in x],
    }

    def __init__(
        self,
        verdict: ReputationVerdict = ReputationVerdict.INCONCLUSIVE,
        category: ReputationCategory = None,
        trust_score: float = None,
        malware_info: MalwareInfo = None,
        verdict_details: List[ReputationTags] = None,
    ):
        self.verdict: ReputationVerdict = verdict
        self.trust_score: float = trust_score
        self.category: Optional[ReputationCategory] = category
        self.malware_info: Optional[MalwareInfo] = malware_info or MalwareInfo()
        self.verdict_details: List[ReputationTags] = verdict_details or list()


class CertificateSummary(Dto):
    encode_settings = {"certificate_status": lambda x: x.name}
    decode_settings = {
        "certificate_status": lambda x: CertificateStatus.UNSIGNED
        if x is None
        else CertificateStatus[x]
    }

    def __init__(
        self,
        certificate_status: CertificateStatus = CertificateStatus.UNSIGNED,
        is_verified: bool = False,
        publisher_name: str = None,
        issuer_name: str = None,
    ):
        self.certificate_status: CertificateStatus = certificate_status
        self.is_verified: bool = is_verified
        self.publisher_name: str = publisher_name
        self.issuer_name: str = issuer_name

    def from_signature_info(self, sig_info: DbSignatureInfo):
        if sig_info:
            self.publisher_name = sig_info.signed_by
            self.issuer_name = sig_info.authority
            if sig_info.signature_status == SignatureStatus.SIGNED_AND_VERIFIED:
                self.is_verified = True
                self.certificate_status = CertificateStatus.SIGNED
            elif sig_info.signature_status in {
                SignatureStatus.SIGNED_NOT_VERIFIED,
                SignatureStatus.INVALID_CERTIFICATE,
            }:
                # don't think invalid certificate is used but it maps here in bit9 db
                self.certificate_status = CertificateStatus.SIGNED
            elif sig_info.signature_status == SignatureStatus.SIGNED_NOT_TRUSTED:
                self.certificate_status = CertificateStatus.REVOKED
            # TODO: no option right now for CertificateStatus.EXPIRED
            #  - but this would be anything signed where the date expired.
            #  The date is not in dynamodb right now.


class SignatureInfo(Dto):
    decode_settings = {
        "certificate_summary": lambda x: CertificateSummary.from_dict(x, True, True)
    }

    def __init__(self, certificate_summary: CertificateSummary = None):
        self.certificate_summary: CertificateSummary = (
            certificate_summary or CertificateSummary()
        )


class FileMetadata(Dto):
    def __init__(
        self,
        sha256: str = None,
        sha1: str = None,
        md5: str = None,
        file_name: str = None,
    ):
        self.sha256: str = sha256
        self.sha1: str = sha1
        self.md5: str = md5
        self.file_name: str = file_name


class ReputationResponse(Dto):
    decode_settings = {
        "classification": lambda x: Classification.from_dict(x, True, True),
        "signature_info": lambda x: SignatureInfo.from_dict(x, True, True),
        "file_metadata": lambda x: FileMetadata.from_dict(x, True, True),
        "last_modified_at": lambda x: None if not x else from_iso_8601(x),
        "first_classified_at": lambda x: None if not x else from_iso_8601(x),
        "expires_at": lambda x: utc_now() if not x else from_iso_8601(x),
    }
    encode_settings = {
        "first_classified_at": lambda x: None if x is None else to_iso_8601(x),
        "last_modified_at": lambda x: None if x is None else to_iso_8601(x),
        "expires_at": lambda x: to_iso_8601(utc_now()) if x is None else to_iso_8601(x),
    }

    def __init__(
        self,
        classification: Classification = None,
        signature_info: SignatureInfo = None,
        file_metadata: FileMetadata = None,
        first_classified_at: datetime = None,
        last_modified_at: datetime = None,
        expires_at: datetime = None,
    ):
        self.classification: Classification = classification or Classification()
        self.signature_info: SignatureInfo = signature_info or SignatureInfo()
        self.file_metadata: FileMetadata = file_metadata or FileMetadata()
        self.first_classified_at: datetime = first_classified_at
        self.last_modified_at: datetime = last_modified_at
        self.expires_at: datetime = expires_at or utc_now()

    def from_reputation_item(self, reputation_item: ReputationItem):
        # start with assigning file details
        self.file_metadata.sha256 = reputation_item.sha256
        self.file_metadata.md5 = reputation_item.md5
        self.file_metadata.sha1 = reputation_item.sha1
        self.file_metadata.file_name = reputation_item.file_name

        # timestamps
        self.last_modified_at = reputation_item.date_modified
        self.first_classified_at = reputation_item.cb_first_seen

        # certificate/signature details
        self.signature_info.certificate_summary.from_signature_info(
            reputation_item.signature_info
        )

        db_reputation = reputation_item.reputation
        self.classification.verdict_details = db_reputation.comment
        if db_reputation.is_locked():
            if db_reputation.trust == 10:
                self.classification.verdict = ReputationVerdict.TRUSTED
            elif db_reputation.threat == 50:
                self.classification.verdict = ReputationVerdict.MALWARE
                if not self.classification.malware_info.group:
                    self.classification.malware_info.group = "PUP"
            elif db_reputation.threat == 100:
                self.classification.verdict = ReputationVerdict.MALWARE

        else:
            # trust/threat assignments
            if db_reputation.trust >= 7:
                self.classification.verdict = ReputationVerdict.TRUSTED
            elif db_reputation.threat >= 50:
                self.classification.verdict = ReputationVerdict.MALWARE
                if db_reputation.threat < 100:
                    self.classification.malware_info.group = "PUP"

            # adaptive/still deciding after trust/threat info
            avira_mfg = False
            if reputation_item.mfg:
                mfg = reputation_item.mfg
                avira_mfg = "Avira" in mfg.classification_scanners
                if self.classification.verdict not in {
                    ReputationVerdict.TRUSTED,
                    ReputationVerdict.SOMEWHAT_TRUSTED,
                }:
                    if avira_mfg:
                        self.classification.verdict = ReputationVerdict.MALWARE
                    elif reputation_item._classification_scanners_count > 2:
                        # TODO: this if/else only matters if we can't have SUSPECT / category PUP.
                        if mfg.malware_group.result in {"Adware", "PUP"}:
                            self.classification.verdict = ReputationVerdict.MALWARE
                        elif (
                            self.classification.verdict
                            == ReputationVerdict.INCONCLUSIVE
                        ):
                            self.classification.verdict = (
                                ReputationVerdict.SUSPECT_MALWARE
                            )

                    # Handle Mimikatz and LaZange PWS case
                    if (
                        reputation_item.mfg.malware_family.result == "Mimikatz"
                        or reputation_item.mfg.malware_family.result == "LaZange"
                    ):
                        if (
                            avira_mfg
                            or reputation_item._classification_scanners_count > 2
                        ):
                            self.classification.malware_info.group = "PWS"
                            reputation_item.mfg.signature = (
                                "PWS:"
                                + (
                                    reputation_item.mfg.platform.result
                                    if reputation_item.mfg.platform.result
                                    else ""
                                )
                                + "/"
                                + (
                                    reputation_item.mfg.malware_family.result
                                    if reputation_item.mfg.malware_family.result
                                    else ""
                                )
                                + (
                                    ("." + ".".join(reputation_item.mfg.tags))
                                    if reputation_item.mfg.tags
                                    else ""
                                )
                            )

            # APC
            if (
                reputation_item.apc
                and not avira_mfg
                and self.classification.verdict != ReputationVerdict.TRUSTED
            ):
                apc = reputation_item.apc
                if (
                    apc.status == ApcStatus.OK
                    and apc.category != ApcCategory.UNKNOWN
                    and apc.category != ApcCategory.CLEAN
                ):
                    self.classification.verdict = ReputationVerdict.MALWARE
                    if apc.verdict == ApcVerdict.PUP:
                        self.classification.malware_info.group = "PUP"

            # still undecided at the end- adaptive or common.
            if (
                self.classification.verdict == ReputationVerdict.INCONCLUSIVE
                and db_reputation.trust > 0
            ):
                self.classification.verdict = ReputationVerdict.SOMEWHAT_TRUSTED

        # if malware/pup and mfg exists, set values.
        if self.classification.verdict not in {
            ReputationVerdict.MALWARE,
            ReputationVerdict.SUSPECT_MALWARE,
        }:
            self.classification.trust_score = db_reputation.trust / 10
        # TODO: how to handle inconclusive but has mfg data?
        elif self.classification.verdict != ReputationVerdict.INCONCLUSIVE:
            if reputation_item.mfg:
                self.classification.malware_info.family = (
                    reputation_item.mfg.malware_family.result
                )
                # malware info needs to retain if pup - example threat 50 or apc PUP.
                self.classification.malware_info.group = (
                    self.classification.malware_info.group
                    or reputation_item.mfg.malware_group.result
                )
                self.classification.malware_info.signature = (
                    reputation_item.mfg.signature
                )
            self.classification.category = ReputationCategory.from_mfg(
                self.classification.malware_info.group
            )
        # set TTL
        self._set_expiration(locked=db_reputation.is_locked())

        return self

    def _set_expiration(self, locked: bool = False) -> None:
        # TODO: what to do with expiration time with only APC conviction?
        now = utc_now()
        seed_week = 7 * random()

        if locked:
            self.expires_at = now + timedelta(days=90 + seed_week)
            return

        elif self.classification.verdict == ReputationVerdict.TRUSTED:
            self.expires_at = now + timedelta(days=60 + seed_week)
            return
        elif self.classification.verdict == ReputationVerdict.SOMEWHAT_TRUSTED:
            self.expires_at = now + timedelta(days=30 + seed_week)
            return
        elif (now - self.first_classified_at).days >= 30:
            self.expires_at = now + timedelta(days=30 + seed_week)
            return

        self.expires_at = now + timedelta(hours=24 + 12 * random())


class ReputationResponseInternal(ReputationResponse):
    def from_reputation_item(self, reputation_item: ReputationItem):
        parent = super().from_reputation_item(reputation_item)
        parent.classification.threat = reputation_item.reputation.threat
        return parent


class ReputationResponseMinimal(ReputationResponse):
    encode_settings = {"category": lambda x: None if not x else x.name}
    decode_settings = {
        "category": lambda x: None if x is None else ReputationCategory[x]
    }

    def __init__(
        self,
        classification: Classification = None,
        signature_info: SignatureInfo = None,
        file_metadata: FileMetadata = None,
        first_classified_at: datetime = None,
        last_modified_at: datetime = None,
        expires_at: datetime = None,
    ):
        super().__init__(
            classification,
            signature_info,
            file_metadata,
            first_classified_at,
            last_modified_at,
            expires_at,
        )
        self.sha256 = file_metadata.sha256 if file_metadata else None
        self.trust = classification.trust_score if classification else None
        self.threat = None
        self.category = classification.category if classification else None

    def from_reputation_item(self, reputation_item: ReputationItem):
        parent = super().from_reputation_item(reputation_item)
        self.sha256 = reputation_item.sha256
        self.trust = parent.classification.trust_score
        self.threat = reputation_item.reputation.threat
        self.category = parent.classification.category
        self.file_metadata = parent.file_metadata
        return self


class ReputationResponseCollection(Dto):
    decode_settings = {
        "responses": lambda x: list()
        if not x
        else [ReputationResponse.from_dict(y, True, True) for y in x],
        "errors": lambda x: list()
        if not x
        else [ReputationErrorResponse.from_dict(y, True, True) for y in x],
    }
    encode_settings = {
        "responses": lambda x: [y.to_dict(True, True) for y in x],
        "errors": lambda x: None if not x else [y.to_dict(True, True) for y in x],
    }

    def __init__(
        self,
        responses: List[ReputationResponse] = None,
        errors: List[ReputationErrorResponse] = None,
    ):
        self.responses: List[ReputationResponse] = responses or list()
        self.errors: List[ReputationErrorResponse] = errors or list()


class ReputationResponseInternalCollection(Dto):
    decode_settings = {
        "responses": lambda x: list()
        if not x
        else [ReputationResponseInternal.from_dict(y, True, True) for y in x],
        "errors": lambda x: list()
        if not x
        else [ReputationErrorResponse.from_dict(y, True, True) for y in x],
    }
    encode_settings = {
        "responses": lambda x: [y.to_dict(True, True) for y in x],
        "errors": lambda x: None if not x else [y.to_dict(True, True) for y in x],
    }

    def __init__(
        self,
        responses: List[ReputationResponseInternal] = None,
        errors: List[ReputationErrorResponse] = None,
    ):
        self.responses: List[ReputationResponseInternal] = responses or list()
        self.errors: List[ReputationErrorResponse] = errors or list()


class ReputationResponseMinimalCollection(Dto):
    decode_settings = {
        "responses": lambda x: list()
        if not x
        else [ReputationResponseMinimal.from_dict(y, True, True) for y in x],
    }
    encode_settings = {
        "responses": lambda x: [
            y.to_dict(
                True,
                True,
                key_ignorelist={
                    "file_metadata",
                    "classification",
                    "expires_at",
                    "first_classified_at",
                    "last_modified_at",
                    "signature_info",
                },
            )
            for y in x
        ]
    }

    def __init__(self, responses: List[ReputationResponseMinimal] = None):
        self.responses: List[ReputationResponseMinimal] = responses or list()
